import os
from typing import Optional
from urllib.parse import urlparse

from gentrace.api_client import ApiClient
from gentrace.apis.tags.v1_api import V1Api
from gentrace.configuration import Configuration

GENTRACE_CONFIG_STATE = {
    "GENTRACE_API_KEY": "",
    "GENTRACE_BASE_PATH": "",
    "GENTRACE_BRANCH": "",
    "GENTRACE_COMMIT": "",
    "GENTRACE_LOG_LEVEL": "warn",
    "global_gentrace_config": None,
    "global_gentrace_api": None,
}


def init(
        api_key: Optional[str] = None,
        host: Optional[str] = None,
        branch: Optional[str] = None,
        commit: Optional[str] = None,
        log_level: Optional[str] = None,
        # @deprecated: use result_name instead
        run_name: Optional[str] = None,
        result_name: Optional[str] = None,
):
    global GENTRACE_CONFIG_STATE

    if not api_key and not os.getenv("GENTRACE_API_KEY"):
        raise ValueError(
            "Gentrace API key was provided neither by the `apiKey` param in the constructor nor by the `GENTRACE_API_KEY` env variable."
        )

    GENTRACE_CONFIG_STATE["GENTRACE_API_KEY"] = api_key or os.getenv("GENTRACE_API_KEY")
    GENTRACE_CONFIG_STATE["GENTRACE_RUN_NAME"] = run_name or os.getenv(
        "GENTRACE_RUN_NAME"
    )

    GENTRACE_CONFIG_STATE["GENTRACE_RESULT_NAME"] = result_name or os.getenv(
        "GENTRACE_RESULT_NAME"
    )

    if host:
        try:
            url = urlparse(host)
            if url.path != "/api" and url.path != "/api/":
                raise ValueError('Gentrace URL path must end in "/api".')
        except Exception as e:
            raise ValueError(f"Invalid Gentrace base path: {str(e)}")

    resolved_host = host if host else "https://gentrace.ai/api"

    GENTRACE_CONFIG_STATE["GENTRACE_BASE_PATH"] = resolved_host

    config = Configuration(host=resolved_host)
    config.access_token = api_key

    GENTRACE_CONFIG_STATE["global_gentrace_config"] = config
    api_client = ApiClient(configuration=config)
    GENTRACE_CONFIG_STATE["global_gentrace_api"] = V1Api(api_client=api_client)

    if branch:
        GENTRACE_CONFIG_STATE["GENTRACE_BRANCH"] = branch

    if commit:
        GENTRACE_CONFIG_STATE["GENTRACE_COMMIT"] = commit

    if log_level:
        if log_level not in ["info", "warn"]:
            raise ValueError("Invalid log level: {}".format(log_level))
        GENTRACE_CONFIG_STATE["GENTRACE_LOG_LEVEL"] = log_level


def deinit():
    global GENTRACE_CONFIG_STATE

    GENTRACE_CONFIG_STATE["GENTRACE_API_KEY"] = ""
    GENTRACE_CONFIG_STATE["GENTRACE_BASE_PATH"] = ""
    GENTRACE_CONFIG_STATE["GENTRACE_BRANCH"] = ""
    GENTRACE_CONFIG_STATE["GENTRACE_COMMIT"] = ""
    GENTRACE_CONFIG_STATE["GENTRACE_LOG_LEVEL"] = "warn"
    GENTRACE_CONFIG_STATE["GENTRACE_RUN_NAME"] = ""
    GENTRACE_CONFIG_STATE["GENTRACE_RESULT_NAME"] = ""

    GENTRACE_CONFIG_STATE["global_gentrace_config"] = None
    GENTRACE_CONFIG_STATE["global_gentrace_api"] = None


__all__ = [
    "init",
    "deinit",
    "GENTRACE_CONFIG_STATE",
]



# coding: utf-8

"""
    Gentrace API

    These API routes are designed to ingest events from clients.  # noqa: E501

    The version of the OpenAPI document: 0.23.0
    Generated by: https://openapi-generator.tech
"""

import decimal  # noqa: F401
import functools  # noqa: F401
import io  # noqa: F401
import re  # noqa: F401
import typing  # noqa: F401
import uuid  # noqa: F401
from datetime import date, datetime  # noqa: F401

import frozendict  # noqa: F401
import typing_extensions  # noqa: F401

from gentrace import schemas  # noqa: F401


class CreateFeedbackV2(
    schemas.DictSchema
):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """


    class MetaOapg:
        required = {
            "score",
            "pipelineRunId",
            "recordedTime",
        }
        
        class properties:
            pipelineRunId = schemas.UUIDSchema
            recordedTime = schemas.Float32Schema
            
            
            class score(
                schemas.Float64Schema
            ):
            
            
                class MetaOapg:
                    format = 'double'
                    inclusive_maximum = 1
                    inclusive_minimum = 0
            
            
            class details(
                schemas.StrBase,
                schemas.NoneBase,
                schemas.Schema,
                schemas.NoneStrMixin
            ):
            
            
                def __new__(
                    cls,
                    *_args: typing.Union[None, str, ],
                    _configuration: typing.Optional[schemas.Configuration] = None,
                ) -> 'details':
                    return super().__new__(
                        cls,
                        *_args,
                        _configuration=_configuration,
                    )
            __annotations__ = {
                "pipelineRunId": pipelineRunId,
                "recordedTime": recordedTime,
                "score": score,
                "details": details,
            }
    
    score: MetaOapg.properties.score
    pipelineRunId: MetaOapg.properties.pipelineRunId
    recordedTime: MetaOapg.properties.recordedTime
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["pipelineRunId"]) -> MetaOapg.properties.pipelineRunId: ...
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["recordedTime"]) -> MetaOapg.properties.recordedTime: ...
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["score"]) -> MetaOapg.properties.score: ...
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["details"]) -> MetaOapg.properties.details: ...
    
    @typing.overload
    def __getitem__(self, name: str) -> schemas.UnsetAnyTypeSchema: ...
    
    def __getitem__(self, name: typing.Union[typing_extensions.Literal["pipelineRunId", "recordedTime", "score", "details", ], str]):
        # dict_instance[name] accessor
        return super().__getitem__(name)
    
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["pipelineRunId"]) -> MetaOapg.properties.pipelineRunId: ...
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["recordedTime"]) -> MetaOapg.properties.recordedTime: ...
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["score"]) -> MetaOapg.properties.score: ...
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["details"]) -> typing.Union[MetaOapg.properties.details, schemas.Unset]: ...
    
    @typing.overload
    def get_item_oapg(self, name: str) -> typing.Union[schemas.UnsetAnyTypeSchema, schemas.Unset]: ...
    
    def get_item_oapg(self, name: typing.Union[typing_extensions.Literal["pipelineRunId", "recordedTime", "score", "details", ], str]):
        return super().get_item_oapg(name)
    

    def __new__(
        cls,
        *_args: typing.Union[dict, frozendict.frozendict, ],
        score: typing.Union[MetaOapg.properties.score, decimal.Decimal, int, float, ],
        pipelineRunId: typing.Union[MetaOapg.properties.pipelineRunId, str, uuid.UUID, ],
        recordedTime: typing.Union[MetaOapg.properties.recordedTime, decimal.Decimal, int, float, ],
        details: typing.Union[MetaOapg.properties.details, None, str, schemas.Unset] = schemas.unset,
        _configuration: typing.Optional[schemas.Configuration] = None,
        **kwargs: typing.Union[schemas.AnyTypeSchema, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, None, list, tuple, bytes],
    ) -> 'CreateFeedbackV2':
        return super().__new__(
            cls,
            *_args,
            score=score,
            pipelineRunId=pipelineRunId,
            recordedTime=recordedTime,
            details=details,
            _configuration=_configuration,
            **kwargs,
        )



from typing import cast

from openai import OpenAI

from gentrace.providers.init import GENTRACE_CONFIG_STATE
from gentrace.providers.llms.openai_v1 import GentraceAsyncOpenAI, GentraceSyncOpenAI


class SimpleGentraceSyncOpenAI(GentraceSyncOpenAI):
    def __init__(self, *args, **kwargs):
        if not GENTRACE_CONFIG_STATE["GENTRACE_API_KEY"]:
            raise ValueError(
                "No Gentrace API key available. Please use init() to set the API key."
            )

        gentrace_config = {
            "api_key": GENTRACE_CONFIG_STATE["GENTRACE_API_KEY"],
            "host": GENTRACE_CONFIG_STATE["GENTRACE_BASE_PATH"],
        }

        super().__init__(*args, **kwargs, gentrace_config=gentrace_config)


SimpleGentraceSyncOpenAITyped = cast(OpenAI, SimpleGentraceSyncOpenAI)


class SimpleGentraceAsyncOpenAI(GentraceAsyncOpenAI):

    def __init__(self, *args, **kwargs):
        if not GENTRACE_CONFIG_STATE["GENTRACE_API_KEY"]:
            raise ValueError(
                "No Gentrace API key available. Please use init() to set the API key."
            )

        gentrace_config = {
            "api_key": GENTRACE_CONFIG_STATE["GENTRACE_API_KEY"],
            "host": GENTRACE_CONFIG_STATE["GENTRACE_BASE_PATH"],
        }

        super().__init__(*args, **kwargs, gentrace_config=gentrace_config)


SimpleGentraceAsyncOpenAITyped = cast(OpenAI, SimpleGentraceAsyncOpenAI)

__all__ = [
    "SimpleGentraceSyncOpenAITyped",
    "SimpleGentraceAsyncOpenAITyped"
]



from typing import Any, Dict, Literal, Optional, TypedDict, Union


class Render(TypedDict):
    type: Literal["html"]
    key: str


class MetadataValueRequired(TypedDict):
    type: str


MetadataValue = Union[MetadataValueRequired, Dict[str, Any]]


class Context(TypedDict):
    userId: Optional[str]
    render: Optional[Render]
    metadata: Optional[Dict[str, MetadataValue]]
    previousRunId: Optional[str]


class ResultContext(TypedDict):
    metadata: Optional[Dict[str, MetadataValue]]


import inspect
from typing import Any, Optional

from gentrace.providers.context import Context
from gentrace.providers.init import GENTRACE_CONFIG_STATE
from gentrace.providers.utils import is_openai_v1


class Pipeline:
    def __init__(
            self,
            # In future releases, we will only support the "slug" parameter. This will no
            # longer be optional.
            slug: Optional[str] = None,
            # @deprecated Use the "slug" parameter instead
            id: Optional[str] = None,
            # @deprecated: use gentrace.providers.init.init() instead to set the Gentrace
            # API key
            api_key: Optional[str] = None,
            # @deprecated: use gentrace.providers.init.init() instead to set the Gentrace
            # base URL
            host: Optional[str] = None,
            openai_config: Any = None,
            pinecone_config: Optional[dict] = None,
    ):
        self.id = id or slug
        self.slug = slug or id

        if api_key:
            self.config = {"api_key": api_key, "host": host}
        else:
            if not GENTRACE_CONFIG_STATE["GENTRACE_API_KEY"]:
                raise ValueError(
                    "No Gentrace API key available. Please use init() to set the API key."
                )
            self.config = {
                "api_key": GENTRACE_CONFIG_STATE["GENTRACE_API_KEY"],
                "host": GENTRACE_CONFIG_STATE["GENTRACE_BASE_PATH"],
            }

        if openai_config:
            try:
                import openai
            except ImportError:
                raise ValueError(
                    "Could not import OpenAI python package. "
                    "Please install it with `pip install openai`."
                )

            if not is_openai_v1():
                for key in openai_config:
                    if key not in openai.__all__:
                        raise ValueError(
                            f"Invalid key ({key}) in supplied OpenAI configuration."
                        )

            self.openai_config = openai_config
        else:
            self.openai_config = None

        if pinecone_config:
            try:
                import pinecone
            except ImportError:
                raise ValueError(
                    "Could not import Pinecone python package. "
                    "Please install it with `pip install pinecone-client`."
                )

            pinecone_init_args = inspect.signature(pinecone.init).parameters.keys()

            for key in pinecone_config:
                if key not in pinecone_init_args:
                    raise ValueError(
                        f"Invalid key ({key}) in supplied Pinecone configuration."
                    )
            self.pinecone_config = pinecone_config
        else:
            self.pinecone_config = None

        self.pipeline_handlers = {}

    def setup(self):
        if self.pinecone_config:
            try:
                from gentrace.providers.vectorstores.pinecone import (
                    PineconePipelineHandler,
                )

                pinecone_handler = PineconePipelineHandler(pipeline=self)
                pinecone_handler.init(api_key=self.pinecone_config["api_key"])
                self.pipeline_handlers["pinecone"] = pinecone_handler
            except ImportError:
                raise ImportError(
                    "Please install Pinecone as a dependency with, e.g. `pip install pinecone-client`"
                )

        if self.openai_config:
            if is_openai_v1():
                from gentrace.providers.llms.openai_v1 import GentraceSyncOpenAI
                openai_handler = GentraceSyncOpenAI(**self.openai_config, gentrace_config=self.config, pipeline=self)
                self.pipeline_handlers["openai"] = openai_handler

            else:
                try:
                    from gentrace.providers.llms.openai_v0 import OpenAIPipelineHandler

                    OpenAIPipelineHandler.setup(self.openai_config)
                    openai_handler = OpenAIPipelineHandler(
                        self.config, self.openai_config, pipeline=self
                    )
                    self.pipeline_handlers["openai"] = openai_handler
                except ImportError:
                    raise ImportError(
                        "Please install OpenAI as a dependency with, e.g. `pip install openai`"
                    )

    def start(self, context: Optional[Context] = None):
        from gentrace.providers.pipeline_run import PipelineRun

        return PipelineRun(pipeline=self, context=context)



from typing import Optional

from gentrace.providers.context import Context


class StepRun:
    def __init__(
        self,
        provider: str,
        invocation: str,
        elapsed_time: float,
        start_time: str,
        end_time: str,
        inputs: any,
        model_params: any,
        outputs: any,
        context: Optional[Context] = None,
    ):
        self.provider = provider
        self.invocation = invocation
        self.elapsed_time = elapsed_time
        self.start_time = start_time
        self.end_time = end_time
        self.inputs = inputs
        self.model_params = model_params
        self.outputs = outputs
        self.context = context or {}



# coding: utf-8

"""
    Gentrace API

    These API routes are designed to ingest events from clients.  # noqa: E501

    The version of the OpenAPI document: 0.23.0
    Generated by: https://openapi-generator.tech
"""

import decimal  # noqa: F401
import functools  # noqa: F401
import io  # noqa: F401
import re  # noqa: F401
import typing  # noqa: F401
import uuid  # noqa: F401
from datetime import date, datetime  # noqa: F401

import frozendict  # noqa: F401
import typing_extensions  # noqa: F401

from gentrace import schemas  # noqa: F401

class SearchableUnixSecondsInput(
    schemas.ComposedSchema,
):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """


    class MetaOapg:
        one_of_0 = schemas.Float32Schema
        
        
        class one_of_1(
            schemas.DictSchema
        ):
        
        
            class MetaOapg:
                
                class properties:
                    gt = schemas.Float32Schema
                    gte = schemas.Float32Schema
                    lt = schemas.Float32Schema
                    lte = schemas.Float32Schema
                    __annotations__ = {
                        "gt": gt,
                        "gte": gte,
                        "lt": lt,
                        "lte": lte,
                    }
            
            @typing.overload
            def __getitem__(self, name: typing_extensions.Literal["gt"]) -> MetaOapg.properties.gt: ...
            
            @typing.overload
            def __getitem__(self, name: typing_extensions.Literal["gte"]) -> MetaOapg.properties.gte: ...
            
            @typing.overload
            def __getitem__(self, name: typing_extensions.Literal["lt"]) -> MetaOapg.properties.lt: ...
            
            @typing.overload
            def __getitem__(self, name: typing_extensions.Literal["lte"]) -> MetaOapg.properties.lte: ...
            
            @typing.overload
            def __getitem__(self, name: str) -> schemas.UnsetAnyTypeSchema: ...
            
            def __getitem__(self, name: typing.Union[typing_extensions.Literal["gt", "gte", "lt", "lte", ], str]):
                # dict_instance[name] accessor
                return super().__getitem__(name)
            
            
            @typing.overload
            def get_item_oapg(self, name: typing_extensions.Literal["gt"]) -> typing.Union[MetaOapg.properties.gt, schemas.Unset]: ...
            
            @typing.overload
            def get_item_oapg(self, name: typing_extensions.Literal["gte"]) -> typing.Union[MetaOapg.properties.gte, schemas.Unset]: ...
            
            @typing.overload
            def get_item_oapg(self, name: typing_extensions.Literal["lt"]) -> typing.Union[MetaOapg.properties.lt, schemas.Unset]: ...
            
            @typing.overload
            def get_item_oapg(self, name: typing_extensions.Literal["lte"]) -> typing.Union[MetaOapg.properties.lte, schemas.Unset]: ...
            
            @typing.overload
            def get_item_oapg(self, name: str) -> typing.Union[schemas.UnsetAnyTypeSchema, schemas.Unset]: ...
            
            def get_item_oapg(self, name: typing.Union[typing_extensions.Literal["gt", "gte", "lt", "lte", ], str]):
                return super().get_item_oapg(name)
            
        
            def __new__(
                cls,
                *_args: typing.Union[dict, frozendict.frozendict, ],
                gt: typing.Union[MetaOapg.properties.gt, decimal.Decimal, int, float, schemas.Unset] = schemas.unset,
                gte: typing.Union[MetaOapg.properties.gte, decimal.Decimal, int, float, schemas.Unset] = schemas.unset,
                lt: typing.Union[MetaOapg.properties.lt, decimal.Decimal, int, float, schemas.Unset] = schemas.unset,
                lte: typing.Union[MetaOapg.properties.lte, decimal.Decimal, int, float, schemas.Unset] = schemas.unset,
                _configuration: typing.Optional[schemas.Configuration] = None,
                **kwargs: typing.Union[schemas.AnyTypeSchema, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, None, list, tuple, bytes],
            ) -> 'one_of_1':
                return super().__new__(
                    cls,
                    *_args,
                    gt=gt,
                    gte=gte,
                    lt=lt,
                    lte=lte,
                    _configuration=_configuration,
                    **kwargs,
                )
        
        @classmethod
        @functools.lru_cache()
        def one_of(cls):
            # we need this here to make our import statements work
            # we must store _composed_schemas in here so the code is only run
            # when we invoke this method. If we kept this at the class
            # level we would get an error because the class level
            # code would be run when this module is imported, and these composed
            # classes don't exist yet because their module has not finished
            # loading
            return [
                cls.one_of_0,
                cls.one_of_1,
            ]


    def __new__(
        cls,
        *_args: typing.Union[dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, bool, None, list, tuple, bytes, io.FileIO, io.BufferedReader, ],
        _configuration: typing.Optional[schemas.Configuration] = None,
        **kwargs: typing.Union[schemas.AnyTypeSchema, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, None, list, tuple, bytes],
    ) -> 'SearchableUnixSecondsInput':
        return super().__new__(
            cls,
            *_args,
            _configuration=_configuration,
            **kwargs,
        )



# coding: utf-8

"""


    Generated by: https://openapi-generator.tech
"""

import decimal  # noqa: F401
import functools  # noqa: F401
import io  # noqa: F401
import re  # noqa: F401
import typing  # noqa: F401
import uuid  # noqa: F401
from dataclasses import dataclass
from datetime import date, datetime  # noqa: F401

import frozendict  # noqa: F401
import typing_extensions  # noqa: F401
import urllib3
from urllib3._collections import HTTPHeaderDict

from gentrace import (
    api_client,
    exceptions,
    schemas,  # noqa: F401
)
from gentrace.model.create_evaluation_v2 import CreateEvaluationV2

from . import path

# body param


class SchemaForRequestBodyApplicationJson(
    schemas.DictSchema
):


    class MetaOapg:
        required = {
            "data",
        }
        
        class properties:
            
            
            class data(
                schemas.ListSchema
            ):
            
            
                class MetaOapg:
                    
                    @staticmethod
                    def items() -> typing.Type['CreateEvaluationV2']:
                        return CreateEvaluationV2
            
                def __new__(
                    cls,
                    _arg: typing.Union[typing.Tuple['CreateEvaluationV2'], typing.List['CreateEvaluationV2']],
                    _configuration: typing.Optional[schemas.Configuration] = None,
                ) -> 'data':
                    return super().__new__(
                        cls,
                        _arg,
                        _configuration=_configuration,
                    )
            
                def __getitem__(self, i: int) -> 'CreateEvaluationV2':
                    return super().__getitem__(i)
            __annotations__ = {
                "data": data,
            }
    
    data: MetaOapg.properties.data
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["data"]) -> MetaOapg.properties.data: ...
    
    @typing.overload
    def __getitem__(self, name: str) -> schemas.UnsetAnyTypeSchema: ...
    
    def __getitem__(self, name: typing.Union[typing_extensions.Literal["data", ], str]):
        # dict_instance[name] accessor
        return super().__getitem__(name)
    
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["data"]) -> MetaOapg.properties.data: ...
    
    @typing.overload
    def get_item_oapg(self, name: str) -> typing.Union[schemas.UnsetAnyTypeSchema, schemas.Unset]: ...
    
    def get_item_oapg(self, name: typing.Union[typing_extensions.Literal["data", ], str]):
        return super().get_item_oapg(name)
    

    def __new__(
        cls,
        *_args: typing.Union[dict, frozendict.frozendict, ],
        data: typing.Union[MetaOapg.properties.data, list, tuple, ],
        _configuration: typing.Optional[schemas.Configuration] = None,
        **kwargs: typing.Union[schemas.AnyTypeSchema, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, None, list, tuple, bytes],
    ) -> 'SchemaForRequestBodyApplicationJson':
        return super().__new__(
            cls,
            *_args,
            data=data,
            _configuration=_configuration,
            **kwargs,
        )


request_body_any_type = api_client.RequestBody(
    content={
        'application/json': api_client.MediaType(
            schema=SchemaForRequestBodyApplicationJson),
    },
    required=True,
)
_auth = [
    'bearerAuth',
]


class SchemaFor200ResponseBodyApplicationJson(
    schemas.DictSchema
):


    class MetaOapg:
        required = {
            "count",
        }
        
        class properties:
            count = schemas.NumberSchema
            __annotations__ = {
                "count": count,
            }
    
    count: MetaOapg.properties.count
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["count"]) -> MetaOapg.properties.count: ...
    
    @typing.overload
    def __getitem__(self, name: str) -> schemas.UnsetAnyTypeSchema: ...
    
    def __getitem__(self, name: typing.Union[typing_extensions.Literal["count", ], str]):
        # dict_instance[name] accessor
        return super().__getitem__(name)
    
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["count"]) -> MetaOapg.properties.count: ...
    
    @typing.overload
    def get_item_oapg(self, name: str) -> typing.Union[schemas.UnsetAnyTypeSchema, schemas.Unset]: ...
    
    def get_item_oapg(self, name: typing.Union[typing_extensions.Literal["count", ], str]):
        return super().get_item_oapg(name)
    

    def __new__(
        cls,
        *_args: typing.Union[dict, frozendict.frozendict, ],
        count: typing.Union[MetaOapg.properties.count, decimal.Decimal, int, float, ],
        _configuration: typing.Optional[schemas.Configuration] = None,
        **kwargs: typing.Union[schemas.AnyTypeSchema, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, None, list, tuple, bytes],
    ) -> 'SchemaFor200ResponseBodyApplicationJson':
        return super().__new__(
            cls,
            *_args,
            count=count,
            _configuration=_configuration,
            **kwargs,
        )


class SchemaFor200ResponseBodyApplicationJsonCharsetutf8(
    schemas.DictSchema
):


    class MetaOapg:
        required = {
            "count",
        }
        
        class properties:
            count = schemas.NumberSchema
            __annotations__ = {
                "count": count,
            }
    
    count: MetaOapg.properties.count
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["count"]) -> MetaOapg.properties.count: ...
    
    @typing.overload
    def __getitem__(self, name: str) -> schemas.UnsetAnyTypeSchema: ...
    
    def __getitem__(self, name: typing.Union[typing_extensions.Literal["count", ], str]):
        # dict_instance[name] accessor
        return super().__getitem__(name)
    
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["count"]) -> MetaOapg.properties.count: ...
    
    @typing.overload
    def get_item_oapg(self, name: str) -> typing.Union[schemas.UnsetAnyTypeSchema, schemas.Unset]: ...
    
    def get_item_oapg(self, name: typing.Union[typing_extensions.Literal["count", ], str]):
        return super().get_item_oapg(name)
    

    def __new__(
        cls,
        *_args: typing.Union[dict, frozendict.frozendict, ],
        count: typing.Union[MetaOapg.properties.count, decimal.Decimal, int, float, ],
        _configuration: typing.Optional[schemas.Configuration] = None,
        **kwargs: typing.Union[schemas.AnyTypeSchema, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, None, list, tuple, bytes],
    ) -> 'SchemaFor200ResponseBodyApplicationJsonCharsetutf8':
        return super().__new__(
            cls,
            *_args,
            count=count,
            _configuration=_configuration,
            **kwargs,
        )


@dataclass
class ApiResponseFor200(api_client.ApiResponse):
    response: urllib3.HTTPResponse
    body: typing.Union[
        SchemaFor200ResponseBodyApplicationJson,
        SchemaFor200ResponseBodyApplicationJsonCharsetutf8,
    ]
    headers: schemas.Unset = schemas.unset


_response_for_200 = api_client.OpenApiResponse(
    response_cls=ApiResponseFor200,
    content={
        'application/json': api_client.MediaType(
            schema=SchemaFor200ResponseBodyApplicationJson),
        'application/json; charset=utf-8': api_client.MediaType(
            schema=SchemaFor200ResponseBodyApplicationJsonCharsetutf8),
    },
)


@dataclass
class ApiResponseFor400(api_client.ApiResponse):
    response: urllib3.HTTPResponse
    body: schemas.Unset = schemas.unset
    headers: schemas.Unset = schemas.unset


_response_for_400 = api_client.OpenApiResponse(
    response_cls=ApiResponseFor400,
)


@dataclass
class ApiResponseFor500(api_client.ApiResponse):
    response: urllib3.HTTPResponse
    body: schemas.Unset = schemas.unset
    headers: schemas.Unset = schemas.unset


_response_for_500 = api_client.OpenApiResponse(
    response_cls=ApiResponseFor500,
)
_status_code_to_response = {
    '200': _response_for_200,
    '400': _response_for_400,
    '500': _response_for_500,
}
_all_accept_content_types = (
    'application/json',
    'application/json; charset=utf-8',
)


class BaseApi(api_client.Api):
    @typing.overload
    def _v2_evaluations_bulk_post_oapg(
        self,
        body: typing.Union[SchemaForRequestBodyApplicationJson,dict, frozendict.frozendict, ],
        content_type: typing_extensions.Literal["application/json"] = ...,
        accept_content_types: typing.Tuple[str] = _all_accept_content_types,
        stream: bool = False,
        timeout: typing.Optional[typing.Union[int, typing.Tuple]] = None,
        skip_deserialization: typing_extensions.Literal[False] = ...,
    ) -> typing.Union[
        ApiResponseFor200,
    ]: ...

    @typing.overload
    def _v2_evaluations_bulk_post_oapg(
        self,
        body: typing.Union[SchemaForRequestBodyApplicationJson,dict, frozendict.frozendict, ],
        content_type: str = ...,
        accept_content_types: typing.Tuple[str] = _all_accept_content_types,
        stream: bool = False,
        timeout: typing.Optional[typing.Union[int, typing.Tuple]] = None,
        skip_deserialization: typing_extensions.Literal[False] = ...,
    ) -> typing.Union[
        ApiResponseFor200,
    ]: ...


    @typing.overload
    def _v2_evaluations_bulk_post_oapg(
        self,
        body: typing.Union[SchemaForRequestBodyApplicationJson,dict, frozendict.frozendict, ],
        skip_deserialization: typing_extensions.Literal[True],
        content_type: str = ...,
        accept_content_types: typing.Tuple[str] = _all_accept_content_types,
        stream: bool = False,
        timeout: typing.Optional[typing.Union[int, typing.Tuple]] = None,
    ) -> api_client.ApiResponseWithoutDeserialization: ...

    @typing.overload
    def _v2_evaluations_bulk_post_oapg(
        self,
        body: typing.Union[SchemaForRequestBodyApplicationJson,dict, frozendict.frozendict, ],
        content_type: str = ...,
        accept_content_types: typing.Tuple[str] = _all_accept_content_types,
        stream: bool = False,
        timeout: typing.Optional[typing.Union[int, typing.Tuple]] = None,
        skip_deserialization: bool = ...,
    ) -> typing.Union[
        ApiResponseFor200,
        api_client.ApiResponseWithoutDeserialization,
    ]: ...

    def _v2_evaluations_bulk_post_oapg(
        self,
        body: typing.Union[SchemaForRequestBodyApplicationJson,dict, frozendict.frozendict, ],
        content_type: str = 'application/json',
        accept_content_types: typing.Tuple[str] = _all_accept_content_types,
        stream: bool = False,
        timeout: typing.Optional[typing.Union[int, typing.Tuple]] = None,
        skip_deserialization: bool = False,
    ):
        """
        Bulk create evaluations
        :param skip_deserialization: If true then api_response.response will be set but
            api_response.body and api_response.headers will not be deserialized into schema
            class instances
        """
        used_path = path.value

        _headers = HTTPHeaderDict()
        # TODO add cookie handling
        if accept_content_types:
            for accept_content_type in accept_content_types:
                _headers.add('Accept', accept_content_type)

        if body is schemas.unset:
            raise exceptions.ApiValueError(
                'The required body parameter has an invalid value of: unset. Set a valid value instead')
        _fields = None
        _body = None
        serialized_data = request_body_any_type.serialize(body, content_type)
        _headers.add('Content-Type', content_type)
        if 'fields' in serialized_data:
            _fields = serialized_data['fields']
        elif 'body' in serialized_data:
            _body = serialized_data['body']
        response = self.api_client.call_api(
            resource_path=used_path,
            method='post'.upper(),
            headers=_headers,
            fields=_fields,
            body=_body,
            auth_settings=_auth,
            stream=stream,
            timeout=timeout,
        )

        if skip_deserialization:
            api_response = api_client.ApiResponseWithoutDeserialization(response=response)
        else:
            response_for_status = _status_code_to_response.get(str(response.status))
            if response_for_status:
                api_response = response_for_status.deserialize(response, self.api_client.configuration)
            else:
                api_response = api_client.ApiResponseWithoutDeserialization(response=response)

        if not 200 <= response.status <= 299:
            raise exceptions.ApiException(
                status=response.status,
                reason=response.reason,
                api_response=api_response
            )

        return api_response


class V2EvaluationsBulkPost(BaseApi):
    # this class is used by api classes that refer to endpoints with operationId fn names

    @typing.overload
    def v2_evaluations_bulk_post(
        self,
        body: typing.Union[SchemaForRequestBodyApplicationJson,dict, frozendict.frozendict, ],
        content_type: typing_extensions.Literal["application/json"] = ...,
        accept_content_types: typing.Tuple[str] = _all_accept_content_types,
        stream: bool = False,
        timeout: typing.Optional[typing.Union[int, typing.Tuple]] = None,
        skip_deserialization: typing_extensions.Literal[False] = ...,
    ) -> typing.Union[
        ApiResponseFor200,
    ]: ...

    @typing.overload
    def v2_evaluations_bulk_post(
        self,
        body: typing.Union[SchemaForRequestBodyApplicationJson,dict, frozendict.frozendict, ],
        content_type: str = ...,
        accept_content_types: typing.Tuple[str] = _all_accept_content_types,
        stream: bool = False,
        timeout: typing.Optional[typing.Union[int, typing.Tuple]] = None,
        skip_deserialization: typing_extensions.Literal[False] = ...,
    ) -> typing.Union[
        ApiResponseFor200,
    ]: ...


    @typing.overload
    def v2_evaluations_bulk_post(
        self,
        body: typing.Union[SchemaForRequestBodyApplicationJson,dict, frozendict.frozendict, ],
        skip_deserialization: typing_extensions.Literal[True],
        content_type: str = ...,
        accept_content_types: typing.Tuple[str] = _all_accept_content_types,
        stream: bool = False,
        timeout: typing.Optional[typing.Union[int, typing.Tuple]] = None,
    ) -> api_client.ApiResponseWithoutDeserialization: ...

    @typing.overload
    def v2_evaluations_bulk_post(
        self,
        body: typing.Union[SchemaForRequestBodyApplicationJson,dict, frozendict.frozendict, ],
        content_type: str = ...,
        accept_content_types: typing.Tuple[str] = _all_accept_content_types,
        stream: bool = False,
        timeout: typing.Optional[typing.Union[int, typing.Tuple]] = None,
        skip_deserialization: bool = ...,
    ) -> typing.Union[
        ApiResponseFor200,
        api_client.ApiResponseWithoutDeserialization,
    ]: ...

    def v2_evaluations_bulk_post(
        self,
        body: typing.Union[SchemaForRequestBodyApplicationJson,dict, frozendict.frozendict, ],
        content_type: str = 'application/json',
        accept_content_types: typing.Tuple[str] = _all_accept_content_types,
        stream: bool = False,
        timeout: typing.Optional[typing.Union[int, typing.Tuple]] = None,
        skip_deserialization: bool = False,
    ):
        return self._v2_evaluations_bulk_post_oapg(
            body=body,
            content_type=content_type,
            accept_content_types=accept_content_types,
            stream=stream,
            timeout=timeout,
            skip_deserialization=skip_deserialization
        )


class ApiForpost(BaseApi):
    # this class is used by api classes that refer to endpoints by path and http method names

    @typing.overload
    def post(
        self,
        body: typing.Union[SchemaForRequestBodyApplicationJson,dict, frozendict.frozendict, ],
        content_type: typing_extensions.Literal["application/json"] = ...,
        accept_content_types: typing.Tuple[str] = _all_accept_content_types,
        stream: bool = False,
        timeout: typing.Optional[typing.Union[int, typing.Tuple]] = None,
        skip_deserialization: typing_extensions.Literal[False] = ...,
    ) -> typing.Union[
        ApiResponseFor200,
    ]: ...

    @typing.overload
    def post(
        self,
        body: typing.Union[SchemaForRequestBodyApplicationJson,dict, frozendict.frozendict, ],
        content_type: str = ...,
        accept_content_types: typing.Tuple[str] = _all_accept_content_types,
        stream: bool = False,
        timeout: typing.Optional[typing.Union[int, typing.Tuple]] = None,
        skip_deserialization: typing_extensions.Literal[False] = ...,
    ) -> typing.Union[
        ApiResponseFor200,
    ]: ...


    @typing.overload
    def post(
        self,
        body: typing.Union[SchemaForRequestBodyApplicationJson,dict, frozendict.frozendict, ],
        skip_deserialization: typing_extensions.Literal[True],
        content_type: str = ...,
        accept_content_types: typing.Tuple[str] = _all_accept_content_types,
        stream: bool = False,
        timeout: typing.Optional[typing.Union[int, typing.Tuple]] = None,
    ) -> api_client.ApiResponseWithoutDeserialization: ...

    @typing.overload
    def post(
        self,
        body: typing.Union[SchemaForRequestBodyApplicationJson,dict, frozendict.frozendict, ],
        content_type: str = ...,
        accept_content_types: typing.Tuple[str] = _all_accept_content_types,
        stream: bool = False,
        timeout: typing.Optional[typing.Union[int, typing.Tuple]] = None,
        skip_deserialization: bool = ...,
    ) -> typing.Union[
        ApiResponseFor200,
        api_client.ApiResponseWithoutDeserialization,
    ]: ...

    def post(
        self,
        body: typing.Union[SchemaForRequestBodyApplicationJson,dict, frozendict.frozendict, ],
        content_type: str = 'application/json',
        accept_content_types: typing.Tuple[str] = _all_accept_content_types,
        stream: bool = False,
        timeout: typing.Optional[typing.Union[int, typing.Tuple]] = None,
        skip_deserialization: bool = False,
    ):
        return self._v2_evaluations_bulk_post_oapg(
            body=body,
            content_type=content_type,
            accept_content_types=accept_content_types,
            stream=stream,
            timeout=timeout,
            skip_deserialization=skip_deserialization
        )





# do not import all endpoints into this module because that uses a lot of memory and stack frames
# if you need the ability to import all endpoints from this module, import them with
# from gentrace.apis.path_to_api import path_to_api

import enum


class PathValues(str, enum.Enum):
    V1_RUN = "/v1/run"
    V1_FEEDBACK = "/v1/feedback"
    V1_TESTCASE = "/v1/test-case"
    V1_TESTRESULT = "/v1/test-result"
    V1_TESTRESULT_ID = "/v1/test-result/{id}"
    V1_TESTRESULT_STATUS = "/v1/test-result/status"
    V1_TESTRESULTSIMPLE = "/v1/test-result-simple"
    V1_PIPELINES = "/v1/pipelines"
    V1_FILES_UPLOAD = "/v1/files/upload"
    V2_FEEDBACK = "/v2/feedback"
    V2_FEEDBACK_ID = "/v2/feedback/{id}"
    V2_TESTRESULTS = "/v2/test-results"
    V2_TESTCASES = "/v2/test-cases"
    V2_TESTCASES_ID = "/v2/test-cases/{id}"
    V2_PIPELINES = "/v2/pipelines"
    V2_FOLDERS = "/v2/folders"
    V2_FOLDERS_ID = "/v2/folders/{id}"
    V2_EVALUATIONS_BULK = "/v2/evaluations/bulk"



# coding: utf-8

"""
    Gentrace API

    These API routes are designed to ingest events from clients.  # noqa: E501

    The version of the OpenAPI document: 0.23.0
    Generated by: https://openapi-generator.tech
"""

import decimal  # noqa: F401
import functools  # noqa: F401
import io  # noqa: F401
import re  # noqa: F401
import typing  # noqa: F401
import uuid  # noqa: F401
from datetime import date, datetime  # noqa: F401

import frozendict  # noqa: F401
import typing_extensions  # noqa: F401

from gentrace import schemas  # noqa: F401

class ExpandedPipeline(
    schemas.ComposedSchema,
):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """


    class MetaOapg:
        
        
        class all_of_1(
            schemas.DictSchema
        ):
        
        
            class MetaOapg:
                
                class properties:
                    
                    
                    class evaluators(
                        schemas.ListSchema
                    ):
                    
                    
                        class MetaOapg:
                            
                            @staticmethod
                            def items() -> typing.Type['TestEvaluator']:
                                return TestEvaluator
                    
                        def __new__(
                            cls,
                            _arg: typing.Union[typing.Tuple['TestEvaluator'], typing.List['TestEvaluator']],
                            _configuration: typing.Optional[schemas.Configuration] = None,
                        ) -> 'evaluators':
                            return super().__new__(
                                cls,
                                _arg,
                                _configuration=_configuration,
                            )
                    
                        def __getitem__(self, i: int) -> 'TestEvaluator':
                            return super().__getitem__(i)
                    __annotations__ = {
                        "evaluators": evaluators,
                    }
            
            @typing.overload
            def __getitem__(self, name: typing_extensions.Literal["evaluators"]) -> MetaOapg.properties.evaluators: ...
            
            @typing.overload
            def __getitem__(self, name: str) -> schemas.UnsetAnyTypeSchema: ...
            
            def __getitem__(self, name: typing.Union[typing_extensions.Literal["evaluators", ], str]):
                # dict_instance[name] accessor
                return super().__getitem__(name)
            
            
            @typing.overload
            def get_item_oapg(self, name: typing_extensions.Literal["evaluators"]) -> typing.Union[MetaOapg.properties.evaluators, schemas.Unset]: ...
            
            @typing.overload
            def get_item_oapg(self, name: str) -> typing.Union[schemas.UnsetAnyTypeSchema, schemas.Unset]: ...
            
            def get_item_oapg(self, name: typing.Union[typing_extensions.Literal["evaluators", ], str]):
                return super().get_item_oapg(name)
            
        
            def __new__(
                cls,
                *_args: typing.Union[dict, frozendict.frozendict, ],
                evaluators: typing.Union[MetaOapg.properties.evaluators, list, tuple, schemas.Unset] = schemas.unset,
                _configuration: typing.Optional[schemas.Configuration] = None,
                **kwargs: typing.Union[schemas.AnyTypeSchema, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, None, list, tuple, bytes],
            ) -> 'all_of_1':
                return super().__new__(
                    cls,
                    *_args,
                    evaluators=evaluators,
                    _configuration=_configuration,
                    **kwargs,
                )
        
        @classmethod
        @functools.lru_cache()
        def all_of(cls):
            # we need this here to make our import statements work
            # we must store _composed_schemas in here so the code is only run
            # when we invoke this method. If we kept this at the class
            # level we would get an error because the class level
            # code would be run when this module is imported, and these composed
            # classes don't exist yet because their module has not finished
            # loading
            return [
                Pipeline,
                cls.all_of_1,
            ]


    def __new__(
        cls,
        *_args: typing.Union[dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, bool, None, list, tuple, bytes, io.FileIO, io.BufferedReader, ],
        _configuration: typing.Optional[schemas.Configuration] = None,
        **kwargs: typing.Union[schemas.AnyTypeSchema, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, None, list, tuple, bytes],
    ) -> 'ExpandedPipeline':
        return super().__new__(
            cls,
            *_args,
            _configuration=_configuration,
            **kwargs,
        )

from gentrace.model.pipeline import Pipeline
from gentrace.model.test_evaluator import TestEvaluator



# coding: utf-8

"""
    Gentrace API

    These API routes are designed to ingest events from clients.  # noqa: E501

    The version of the OpenAPI document: 0.23.0
    Generated by: https://openapi-generator.tech
"""

import decimal  # noqa: F401
import functools  # noqa: F401
import io  # noqa: F401
import re  # noqa: F401
import typing  # noqa: F401
import uuid  # noqa: F401
from datetime import date, datetime  # noqa: F401

import frozendict  # noqa: F401
import typing_extensions  # noqa: F401

from gentrace import schemas  # noqa: F401

class FeedbackResponse(
    schemas.DictSchema
):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """


    class MetaOapg:
        required = {
            "message",
        }
        
        class properties:
            message = schemas.StrSchema
            __annotations__ = {
                "message": message,
            }
        additional_properties = schemas.NotAnyTypeSchema
    
    message: MetaOapg.properties.message
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["message"]) -> MetaOapg.properties.message: ...
    
    def __getitem__(self, name: typing.Union[typing_extensions.Literal["message"], ]):
        # dict_instance[name] accessor
        return super().__getitem__(name)
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["message"]) -> MetaOapg.properties.message: ...
    
    def get_item_oapg(self, name: typing.Union[typing_extensions.Literal["message"], ]):
        return super().get_item_oapg(name)

    def __new__(
        cls,
        *_args: typing.Union[dict, frozendict.frozendict, ],
        message: typing.Union[MetaOapg.properties.message, str, ],
        _configuration: typing.Optional[schemas.Configuration] = None,
    ) -> 'FeedbackResponse':
        return super().__new__(
            cls,
            *_args,
            message=message,
            _configuration=_configuration,
        )



# coding: utf-8

# flake8: noqa

# import all models into this package
# if you have many models here with many references from one model to another this may
# raise a RecursionError
# to avoid this, import only the models that you directly need like:
# from gentrace.model.pet import Pet
# or import this package, but before doing it, use:
# import sys
# sys.setrecursionlimit(n)

from gentrace.model.create_evaluation_v2 import CreateEvaluationV2
from gentrace.model.create_feedback_v2 import CreateFeedbackV2
from gentrace.model.create_multiple_test_cases import CreateMultipleTestCases
from gentrace.model.create_single_test_case import CreateSingleTestCase
from gentrace.model.expanded_pipeline import ExpandedPipeline
from gentrace.model.expanded_test_result import ExpandedTestResult
from gentrace.model.expanded_test_run import ExpandedTestRun
from gentrace.model.feedback_request import FeedbackRequest
from gentrace.model.feedback_response import FeedbackResponse
from gentrace.model.feedback_v2 import FeedbackV2
from gentrace.model.filterable_metadata_input import FilterableMetadataInput
from gentrace.model.folder_v2 import FolderV2
from gentrace.model.full_run import FullRun
from gentrace.model.metadata_value_object import MetadataValueObject
from gentrace.model.pipeline import Pipeline
from gentrace.model.pipeline_v2 import PipelineV2
from gentrace.model.resolved_step_run import ResolvedStepRun
from gentrace.model.run_request import RunRequest
from gentrace.model.run_response import RunResponse
from gentrace.model.searchable_string_input import SearchableStringInput
from gentrace.model.searchable_unix_seconds_input import SearchableUnixSecondsInput
from gentrace.model.step_run import StepRun
from gentrace.model.test_case import TestCase
from gentrace.model.test_case_v2 import TestCaseV2
from gentrace.model.test_evaluation import TestEvaluation
from gentrace.model.test_evaluator import TestEvaluator
from gentrace.model.test_result import TestResult
from gentrace.model.test_result_v2 import TestResultV2
from gentrace.model.test_run import TestRun
from gentrace.model.unix_seconds import UnixSeconds
from gentrace.model.update_feedback_v2 import UpdateFeedbackV2
from gentrace.model.update_test_case import UpdateTestCase



import asyncio
import logging
import re
from concurrent.futures import ThreadPoolExecutor
from datetime import datetime, timezone
from importlib.metadata import version

from gentrace.apis.tags.v1_api import V1Api
from gentrace.models import RunRequest
from gentrace.providers.init import GENTRACE_CONFIG_STATE

__all__ = [
    "to_date_string",
    "from_date_string",
    "run_post_background",
    "log_debug",
    "log_info",
    "log_warn",
    "get_test_counter",
    "increment_test_counter",
    "decrement_test_counter",
    "is_openai_v1"
]

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

# Create a stream handler that writes to stderr
stdout_handler = logging.StreamHandler()
logger.setLevel(logging.INFO)

formatter = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
stdout_handler.setFormatter(formatter)

# Add the handlers to the logger
logger.addHandler(stdout_handler)


def to_date_string(time_value):
    utc_time = datetime.utcfromtimestamp(time_value)
    utc_time_str = utc_time.strftime("%Y-%m-%dT%H:%M:%S.%fZ")
    return utc_time_str[:-4] + "Z"


def from_date_string(time_stamp):
    # If the string ends with 'Z', remove it and set it as UTC.
    if time_stamp.endswith('Z'):
        time_stamp = time_stamp[:-1]
        dt = datetime.fromisoformat(time_stamp).replace(tzinfo=timezone.utc)
    else:
        dt = datetime.fromisoformat(time_stamp)

    timestamp = dt.timestamp()
    seconds_since_epoch = int(timestamp)
    return seconds_since_epoch


def is_openai_v1():
    version_info = version("openai")
    if version_info.startswith("1."):
        return True
    return False


def log_debug(message, **params):
    log_level = GENTRACE_CONFIG_STATE["GENTRACE_LOG_LEVEL"]
    if not log_level:
        return

    library_log_level = logging.INFO if log_level == "info" else logging.WARN
    logger.setLevel(library_log_level)

    msg = logfmt(dict(message=message, **params))
    logger.debug(msg)


def log_info(message, **params):
    log_level = GENTRACE_CONFIG_STATE["GENTRACE_LOG_LEVEL"]
    if not log_level:
        return

    library_log_level = logging.INFO if log_level == "info" else logging.WARN
    logger.setLevel(library_log_level)

    msg = logfmt(dict(message=message, **params))
    logger.info(msg)


def log_warn(message, **params):
    log_level = GENTRACE_CONFIG_STATE["GENTRACE_LOG_LEVEL"]
    if not log_level:
        return

    library_log_level = logging.INFO if log_level == "info" else logging.WARN
    logger.setLevel(library_log_level)

    msg = logfmt(dict(message=message, **params))
    logger.warn(msg)


# Logger exception automatically logs the stack trace
def log_exception(message, **params):
    log_level = GENTRACE_CONFIG_STATE["GENTRACE_LOG_LEVEL"]
    if not log_level:
        return

    library_log_level = logging.INFO if log_level == "info" else logging.WARN
    logger.setLevel(library_log_level)

    logger.exception(message)


def logfmt(props):
    def fmt(key, val):
        # Handle case where val is a bytes or bytesarray
        if hasattr(val, "decode"):
            val = val.decode("utf-8")
        # Check if val is already a string to avoid re-encoding into ascii.
        if not isinstance(val, str):
            val = str(val)
        if re.search(r"\s", val):
            val = repr(val)
        # key should already be a string
        if re.search(r"\s", key):
            key = repr(key)
        return "{key}={val}".format(key=key, val=val)

    return " ".join([fmt(key, val) for key, val in sorted(props.items())])


async def run_post_background(api_instance: V1Api, pipeline_run_data: RunRequest):
    def wrapped_api_invocation():
        try:
            log_info("Submitting PipelineRun to Gentrace")
            result = api_instance.v1_run_post(pipeline_run_data)
            log_info("Successfully submitted PipelineRun to Gentrace")
            return result
        except Exception as e:
            log_exception("Error submitting to Gentrace: ")
            return None

    with ThreadPoolExecutor() as executor:
        result = await asyncio.get_event_loop().run_in_executor(
            executor, wrapped_api_invocation
        )
    return result


test_counter = 0


def get_test_counter():
    return test_counter


def increment_test_counter():
    global test_counter
    test_counter += 1
    return test_counter


def decrement_test_counter():
    global test_counter
    test_counter -= 1
    return test_counter



# coding: utf-8

"""
    Gentrace API

    These API routes are designed to ingest events from clients.  # noqa: E501

    The version of the OpenAPI document: 0.23.0
    Generated by: https://openapi-generator.tech
"""

import decimal  # noqa: F401
import functools  # noqa: F401
import io  # noqa: F401
import re  # noqa: F401
import typing  # noqa: F401
import uuid  # noqa: F401
from datetime import date, datetime  # noqa: F401

import frozendict  # noqa: F401
import typing_extensions  # noqa: F401

from gentrace import schemas  # noqa: F401

class FilterableMetadataInput(
    schemas.DictSchema
):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.

    Metadata filters look like { prompt: { contains: "My prompt" } }
    """


    class MetaOapg:
        
        
        class additional_properties(
            schemas.DictSchema
        ):
        
        
            class MetaOapg:
                
                class properties:
                    exists = schemas.BoolSchema
                    contains = schemas.StrSchema
                    
                    
                    class equals(
                        schemas.ComposedSchema,
                    ):
                    
                    
                        class MetaOapg:
                            one_of_0 = schemas.StrSchema
                            one_of_1 = schemas.NumberSchema
                            
                            @classmethod
                            @functools.lru_cache()
                            def one_of(cls):
                                # we need this here to make our import statements work
                                # we must store _composed_schemas in here so the code is only run
                                # when we invoke this method. If we kept this at the class
                                # level we would get an error because the class level
                                # code would be run when this module is imported, and these composed
                                # classes don't exist yet because their module has not finished
                                # loading
                                return [
                                    cls.one_of_0,
                                    cls.one_of_1,
                                ]
                    
                    
                        def __new__(
                            cls,
                            *_args: typing.Union[dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, bool, None, list, tuple, bytes, io.FileIO, io.BufferedReader, ],
                            _configuration: typing.Optional[schemas.Configuration] = None,
                            **kwargs: typing.Union[schemas.AnyTypeSchema, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, None, list, tuple, bytes],
                        ) -> 'equals':
                            return super().__new__(
                                cls,
                                *_args,
                                _configuration=_configuration,
                                **kwargs,
                            )
                    
                    
                    class gt(
                        schemas.ComposedSchema,
                    ):
                    
                    
                        class MetaOapg:
                            one_of_0 = schemas.StrSchema
                            one_of_1 = schemas.NumberSchema
                            
                            @classmethod
                            @functools.lru_cache()
                            def one_of(cls):
                                # we need this here to make our import statements work
                                # we must store _composed_schemas in here so the code is only run
                                # when we invoke this method. If we kept this at the class
                                # level we would get an error because the class level
                                # code would be run when this module is imported, and these composed
                                # classes don't exist yet because their module has not finished
                                # loading
                                return [
                                    cls.one_of_0,
                                    cls.one_of_1,
                                ]
                    
                    
                        def __new__(
                            cls,
                            *_args: typing.Union[dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, bool, None, list, tuple, bytes, io.FileIO, io.BufferedReader, ],
                            _configuration: typing.Optional[schemas.Configuration] = None,
                            **kwargs: typing.Union[schemas.AnyTypeSchema, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, None, list, tuple, bytes],
                        ) -> 'gt':
                            return super().__new__(
                                cls,
                                *_args,
                                _configuration=_configuration,
                                **kwargs,
                            )
                    
                    
                    class gte(
                        schemas.ComposedSchema,
                    ):
                    
                    
                        class MetaOapg:
                            one_of_0 = schemas.StrSchema
                            one_of_1 = schemas.NumberSchema
                            
                            @classmethod
                            @functools.lru_cache()
                            def one_of(cls):
                                # we need this here to make our import statements work
                                # we must store _composed_schemas in here so the code is only run
                                # when we invoke this method. If we kept this at the class
                                # level we would get an error because the class level
                                # code would be run when this module is imported, and these composed
                                # classes don't exist yet because their module has not finished
                                # loading
                                return [
                                    cls.one_of_0,
                                    cls.one_of_1,
                                ]
                    
                    
                        def __new__(
                            cls,
                            *_args: typing.Union[dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, bool, None, list, tuple, bytes, io.FileIO, io.BufferedReader, ],
                            _configuration: typing.Optional[schemas.Configuration] = None,
                            **kwargs: typing.Union[schemas.AnyTypeSchema, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, None, list, tuple, bytes],
                        ) -> 'gte':
                            return super().__new__(
                                cls,
                                *_args,
                                _configuration=_configuration,
                                **kwargs,
                            )
                    
                    
                    class lt(
                        schemas.ComposedSchema,
                    ):
                    
                    
                        class MetaOapg:
                            one_of_0 = schemas.StrSchema
                            one_of_1 = schemas.NumberSchema
                            
                            @classmethod
                            @functools.lru_cache()
                            def one_of(cls):
                                # we need this here to make our import statements work
                                # we must store _composed_schemas in here so the code is only run
                                # when we invoke this method. If we kept this at the class
                                # level we would get an error because the class level
                                # code would be run when this module is imported, and these composed
                                # classes don't exist yet because their module has not finished
                                # loading
                                return [
                                    cls.one_of_0,
                                    cls.one_of_1,
                                ]
                    
                    
                        def __new__(
                            cls,
                            *_args: typing.Union[dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, bool, None, list, tuple, bytes, io.FileIO, io.BufferedReader, ],
                            _configuration: typing.Optional[schemas.Configuration] = None,
                            **kwargs: typing.Union[schemas.AnyTypeSchema, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, None, list, tuple, bytes],
                        ) -> 'lt':
                            return super().__new__(
                                cls,
                                *_args,
                                _configuration=_configuration,
                                **kwargs,
                            )
                    
                    
                    class lte(
                        schemas.ComposedSchema,
                    ):
                    
                    
                        class MetaOapg:
                            one_of_0 = schemas.StrSchema
                            one_of_1 = schemas.NumberSchema
                            
                            @classmethod
                            @functools.lru_cache()
                            def one_of(cls):
                                # we need this here to make our import statements work
                                # we must store _composed_schemas in here so the code is only run
                                # when we invoke this method. If we kept this at the class
                                # level we would get an error because the class level
                                # code would be run when this module is imported, and these composed
                                # classes don't exist yet because their module has not finished
                                # loading
                                return [
                                    cls.one_of_0,
                                    cls.one_of_1,
                                ]
                    
                    
                        def __new__(
                            cls,
                            *_args: typing.Union[dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, bool, None, list, tuple, bytes, io.FileIO, io.BufferedReader, ],
                            _configuration: typing.Optional[schemas.Configuration] = None,
                            **kwargs: typing.Union[schemas.AnyTypeSchema, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, None, list, tuple, bytes],
                        ) -> 'lte':
                            return super().__new__(
                                cls,
                                *_args,
                                _configuration=_configuration,
                                **kwargs,
                            )
                    __annotations__ = {
                        "exists": exists,
                        "contains": contains,
                        "equals": equals,
                        "gt": gt,
                        "gte": gte,
                        "lt": lt,
                        "lte": lte,
                    }
            
            @typing.overload
            def __getitem__(self, name: typing_extensions.Literal["exists"]) -> MetaOapg.properties.exists: ...
            
            @typing.overload
            def __getitem__(self, name: typing_extensions.Literal["contains"]) -> MetaOapg.properties.contains: ...
            
            @typing.overload
            def __getitem__(self, name: typing_extensions.Literal["equals"]) -> MetaOapg.properties.equals: ...
            
            @typing.overload
            def __getitem__(self, name: typing_extensions.Literal["gt"]) -> MetaOapg.properties.gt: ...
            
            @typing.overload
            def __getitem__(self, name: typing_extensions.Literal["gte"]) -> MetaOapg.properties.gte: ...
            
            @typing.overload
            def __getitem__(self, name: typing_extensions.Literal["lt"]) -> MetaOapg.properties.lt: ...
            
            @typing.overload
            def __getitem__(self, name: typing_extensions.Literal["lte"]) -> MetaOapg.properties.lte: ...
            
            @typing.overload
            def __getitem__(self, name: str) -> schemas.UnsetAnyTypeSchema: ...
            
            def __getitem__(self, name: typing.Union[typing_extensions.Literal["exists", "contains", "equals", "gt", "gte", "lt", "lte", ], str]):
                # dict_instance[name] accessor
                return super().__getitem__(name)
            
            
            @typing.overload
            def get_item_oapg(self, name: typing_extensions.Literal["exists"]) -> typing.Union[MetaOapg.properties.exists, schemas.Unset]: ...
            
            @typing.overload
            def get_item_oapg(self, name: typing_extensions.Literal["contains"]) -> typing.Union[MetaOapg.properties.contains, schemas.Unset]: ...
            
            @typing.overload
            def get_item_oapg(self, name: typing_extensions.Literal["equals"]) -> typing.Union[MetaOapg.properties.equals, schemas.Unset]: ...
            
            @typing.overload
            def get_item_oapg(self, name: typing_extensions.Literal["gt"]) -> typing.Union[MetaOapg.properties.gt, schemas.Unset]: ...
            
            @typing.overload
            def get_item_oapg(self, name: typing_extensions.Literal["gte"]) -> typing.Union[MetaOapg.properties.gte, schemas.Unset]: ...
            
            @typing.overload
            def get_item_oapg(self, name: typing_extensions.Literal["lt"]) -> typing.Union[MetaOapg.properties.lt, schemas.Unset]: ...
            
            @typing.overload
            def get_item_oapg(self, name: typing_extensions.Literal["lte"]) -> typing.Union[MetaOapg.properties.lte, schemas.Unset]: ...
            
            @typing.overload
            def get_item_oapg(self, name: str) -> typing.Union[schemas.UnsetAnyTypeSchema, schemas.Unset]: ...
            
            def get_item_oapg(self, name: typing.Union[typing_extensions.Literal["exists", "contains", "equals", "gt", "gte", "lt", "lte", ], str]):
                return super().get_item_oapg(name)
            
        
            def __new__(
                cls,
                *_args: typing.Union[dict, frozendict.frozendict, ],
                exists: typing.Union[MetaOapg.properties.exists, bool, schemas.Unset] = schemas.unset,
                contains: typing.Union[MetaOapg.properties.contains, str, schemas.Unset] = schemas.unset,
                equals: typing.Union[MetaOapg.properties.equals, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, bool, None, list, tuple, bytes, io.FileIO, io.BufferedReader, schemas.Unset] = schemas.unset,
                gt: typing.Union[MetaOapg.properties.gt, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, bool, None, list, tuple, bytes, io.FileIO, io.BufferedReader, schemas.Unset] = schemas.unset,
                gte: typing.Union[MetaOapg.properties.gte, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, bool, None, list, tuple, bytes, io.FileIO, io.BufferedReader, schemas.Unset] = schemas.unset,
                lt: typing.Union[MetaOapg.properties.lt, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, bool, None, list, tuple, bytes, io.FileIO, io.BufferedReader, schemas.Unset] = schemas.unset,
                lte: typing.Union[MetaOapg.properties.lte, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, bool, None, list, tuple, bytes, io.FileIO, io.BufferedReader, schemas.Unset] = schemas.unset,
                _configuration: typing.Optional[schemas.Configuration] = None,
                **kwargs: typing.Union[schemas.AnyTypeSchema, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, None, list, tuple, bytes],
            ) -> 'additional_properties':
                return super().__new__(
                    cls,
                    *_args,
                    exists=exists,
                    contains=contains,
                    equals=equals,
                    gt=gt,
                    gte=gte,
                    lt=lt,
                    lte=lte,
                    _configuration=_configuration,
                    **kwargs,
                )
    
    def __getitem__(self, name: typing.Union[str, ]) -> MetaOapg.additional_properties:
        # dict_instance[name] accessor
        return super().__getitem__(name)
    
    def get_item_oapg(self, name: typing.Union[str, ]) -> MetaOapg.additional_properties:
        return super().get_item_oapg(name)

    def __new__(
        cls,
        *_args: typing.Union[dict, frozendict.frozendict, ],
        _configuration: typing.Optional[schemas.Configuration] = None,
        **kwargs: typing.Union[MetaOapg.additional_properties, dict, frozendict.frozendict, ],
    ) -> 'FilterableMetadataInput':
        return super().__new__(
            cls,
            *_args,
            _configuration=_configuration,
            **kwargs,
        )



import io  # noqa: F401
import os
import tempfile
import typing

from gentrace.api_client import ApiClient
from gentrace.apis.tags.v1_api import V1Api
from gentrace.paths.v1_files_upload.post import SchemaForRequestBodyMultipartFormData
from gentrace.providers.init import (
    GENTRACE_CONFIG_STATE,
)
from gentrace.schemas import BinarySchema


def upload_file(
        file: typing.Union[BinarySchema, bytes, io.FileIO, io.BufferedReader]
) -> str:
    """Uploads a file to the Gentrace API.

    Parameters:
    - file (typing.Union[BinarySchema, bytes, io.FileIO, io.BufferedReader]): The file to be uploaded.

    Returns:
    - str: URL of the uploaded file.

    Raises:
    - ValueError: If the Gentrace API key is not initialized

    Note:
    Ensure that the Gentrace API is initialized by calling init() before using this function.
    """
    config = GENTRACE_CONFIG_STATE["global_gentrace_config"]
    if not config:
        raise ValueError("Gentrace API key not initialized. Call init() first.")

    api_client = ApiClient(configuration=config)
    api = V1Api(api_client=api_client)

    body = SchemaForRequestBodyMultipartFormData(file=file)
    response = api.v1_files_upload_post(body=body)
    url = response.body.get("url", None)
    return url


def upload_bytes(file_name: str, content: bytes):
    """Uploads a file as bytes to the Gentrace API.

    Parameters:
    - file_name (str): The name of the file to be uploaded.
    - content (bytes): The content of the file to be uploaded.

    Returns:
    - str: URL of the uploaded file.

    Raises:
    - ValueError: If the Gentrace API key is not initialized

    Note:
    Ensure that the Gentrace API is initialized by calling init() before using this function.
    """
    config = GENTRACE_CONFIG_STATE["global_gentrace_config"]
    if not config:
        raise ValueError("Gentrace API key not initialized. Call init() first.")

    temp_dir = tempfile.gettempdir()
    temp_file_path = os.path.join(temp_dir, file_name)

    with open(temp_file_path, "wb") as temp:
        temp.write(content)

    file = io.FileIO(temp_file_path, mode="rb")

    url = upload_file(file)

    file.close()
    os.remove(temp_file_path)

    return url



# coding: utf-8

"""
    Gentrace API

    These API routes are designed to ingest events from clients.  # noqa: E501

    The version of the OpenAPI document: 0.23.0
    Generated by: https://openapi-generator.tech
"""

import decimal  # noqa: F401
import functools  # noqa: F401
import io  # noqa: F401
import re  # noqa: F401
import typing  # noqa: F401
import uuid  # noqa: F401
from datetime import date, datetime  # noqa: F401

import frozendict  # noqa: F401
import typing_extensions  # noqa: F401

from gentrace import schemas  # noqa: F401

UnixSeconds = schemas.Float32Schema



import copy
import os
import time
import uuid
from typing import Any, Dict, Optional

import pystache
from openai import AsyncOpenAI, OpenAI, chat, completions, embeddings

from gentrace.providers.pipeline import Pipeline
from gentrace.providers.pipeline_run import PipelineRun
from gentrace.providers.step_run import StepRun
from gentrace.providers.utils import to_date_string

# Mega hack: somehow __class__ is actually invoking the OpenAI constructor
prior_value = os.environ.get("OPENAI_API_KEY", None)

os.environ["OPENAI_API_KEY"] = "DUMMY_VALUE"

# Extract classes as you mentioned
ExtractedCompletions = completions.__class__
ExtractedChat = chat.__class__
ExtractedChatCompletions = chat.completions.__class__
ExtractedEmbeddings = embeddings.__class__

if prior_value is not None:
    os.environ["OPENAI_API_KEY"] = prior_value
else:
    del os.environ["OPENAI_API_KEY"]

# Hack: dummy async OpenAI client to get async classes
dummy_client = AsyncOpenAI(api_key="DUMMY_VALUE")

ExtractedAsyncCompletions = dummy_client.completions.__class__
ExtractedAsyncChat = dummy_client.chat.__class__
ExtractedAsyncChatCompletions = dummy_client.chat.completions.__class__
ExtractedAsyncEmbeddings = dummy_client.embeddings.__class__


class GentraceAsyncOpenAI(AsyncOpenAI):
    pipeline_run: Optional[PipelineRun] = None

    def __init__(self, *args, **kwargs):
        # print all kwargs
        self.pipeline_run = kwargs.pop("pipeline_run", None)
        self.pipeline = kwargs.pop("pipeline", None)
        self.config = kwargs.pop("gentrace_config")

        super().__init__(*args, **kwargs)

        self.completions = GentraceAsyncCompletions(client=self, gentrace_config=self.config,
                                                    pipeline=self.pipeline,
                                                    pipeline_run=self.pipeline_run)
        self.chat = ExtractedAsyncChat(self)
        self.chat.completions = GentraceAsyncChatCompletions(client=self, gentrace_config=self.config,
                                                             pipeline=self.pipeline,
                                                             pipeline_run=self.pipeline_run)
        self.embeddings = GentraceAsyncEmbeddings(client=self, gentrace_config=self.config,
                                                  pipeline=self.pipeline,
                                                  pipeline_run=self.pipeline_run)


class GentraceAsyncEmbeddings(ExtractedAsyncEmbeddings):

    def __init__(self, *, gentrace_config: Dict[str, Any], client: AsyncOpenAI, pipeline: Optional[Pipeline] = None,
                 pipeline_run: Optional[PipelineRun] = None):
        super().__init__(client)

        self.pipeline = pipeline
        self.pipeline_run = pipeline_run
        self.gentrace_config = gentrace_config

    async def create(self, *args, **kwargs):
        model = kwargs.get("model")
        pipeline_slug = kwargs.pop("pipeline_slug", None)

        context = kwargs.pop("gentrace", {})
        input_params = {k: v for k, v in kwargs.items() if k not in ["model"]}

        start_time = time.time()
        completion = await super().create(*args, **kwargs)
        end_time = time.time()

        elapsed_time = int((end_time - start_time) * 1000)

        pipeline_run = self.pipeline_run

        is_self_contained = not pipeline_run and pipeline_slug

        if is_self_contained:
            pipeline = Pipeline(
                slug=pipeline_slug,
                **self.gentrace_config,
            )

            pipeline_run = PipelineRun(pipeline=pipeline, context=context)

        if pipeline_run:
            pipeline_run.add_step_run(
                OpenAICreateEmbeddingStepRun(
                    elapsed_time,
                    to_date_string(start_time),
                    to_date_string(end_time),
                    input_params,
                    {"model": model},
                    completion.dict(),
                    context,
                )
            )

            if is_self_contained:
                submit_result = pipeline_run.submit()
                setattr(completion, "pipelineRunId",
                        submit_result["pipelineRunId"] if "pipelineRunId" in submit_result else None)

        return completion


class GentraceAsyncCompletions(ExtractedAsyncCompletions):
    def __init__(self, *, gentrace_config: Dict[str, Any], client: AsyncOpenAI, pipeline: Optional[Pipeline] = None,
                 pipeline_run: Optional[PipelineRun] = None):
        super().__init__(client)

        self.pipeline = pipeline
        self.pipeline_run = pipeline_run
        self.gentrace_config = gentrace_config

    async def create(self, *args, **kwargs):
        prompt_template = kwargs.get("prompt_template")
        prompt_inputs = kwargs.get("prompt_inputs")
        prompt = kwargs.get("prompt")
        pipeline_slug = kwargs.pop("pipeline_slug", None)

        context = kwargs.pop("gentrace", {})
        stream = kwargs.get("stream")
        base_completion_options = {
            k: v
            for k, v in kwargs.items()
            if k not in ["prompt_template", "prompt_inputs"]
        }

        if stream:
            rendered_prompt = prompt

            if prompt_template and prompt_inputs:
                rendered_prompt = pystache.render(prompt_template, prompt_inputs)
            else:
                # If no prompt template or inputs, then specify the prompt as the entire input
                prompt_inputs = prompt

            new_completion_options = {
                **base_completion_options,
                "prompt": rendered_prompt,
            }

            start_time = time.time()
            completion = await super().create(**new_completion_options)

            is_self_contained = (
                    not self.pipeline_run and pipeline_slug
            )
            if is_self_contained:
                pipeline_run_id = str(uuid.uuid4())

            async def profiled_completion():
                modified_response = []
                async for value in completion:
                    if value and is_self_contained:
                        setattr(value, "pipelineRunId", pipeline_run_id)
                    modified_response.append(value)
                    yield value

                end_time = time.time()

                full_response = create_completion_stream_response(modified_response)

                create_completion_step_run(
                    self.pipeline_run,
                    pipeline_slug,
                    self.gentrace_config,
                    start_time,
                    end_time,
                    base_completion_options,
                    prompt_template,
                    prompt_inputs,
                    completion,
                    full_response,
                    pipeline_run_id if is_self_contained else None,
                    stream,
                    context,
                )

            return profiled_completion()

        rendered_prompt = prompt

        if prompt_template and prompt_inputs:
            rendered_prompt = pystache.render(prompt_template, prompt_inputs)
        else:
            # If no prompt template or inputs, then specify the prompt as the entire input
            prompt_inputs = prompt

        new_completion_options = {
            **base_completion_options,
            "prompt": rendered_prompt,
        }

        start_time = time.time()
        completion = await super().create(**new_completion_options)
        end_time = time.time()

        create_completion_step_run(
            self.pipeline_run,
            pipeline_slug,
            self.gentrace_config,
            start_time,
            end_time,
            base_completion_options,
            prompt_template,
            prompt_inputs,
            completion,
            completion.dict(),
            None,
            stream,
            context,
        )
        return completion


class GentraceAsyncChatCompletions(ExtractedAsyncChatCompletions):
    def __init__(self, *, gentrace_config: Dict[str, Any], client: AsyncOpenAI, pipeline: Optional[Pipeline] = None,
                 pipeline_run: Optional[PipelineRun] = None):
        super().__init__(client)

        self.pipeline = pipeline
        self.pipeline_run = pipeline_run
        self.gentrace_config = gentrace_config

    async def create(self, *args, **kwargs):
        messages = kwargs.get("messages")
        user = kwargs.get("user")
        stream = kwargs.get("stream")
        pipeline_slug = kwargs.pop("pipeline_slug", None)
        context = kwargs.pop("gentrace", {})
        model_params = {
            k: v for k, v in kwargs.items() if k not in ["messages", "user"]
        }

        content_templates_array = [
            message["contentTemplate"] if "contentTemplate" in message else None
            for message in messages
        ]

        content_inputs_array = [
            message["contentInputs"] if "contentInputs" in message else None
            for message in messages
        ]

        rendered_messages = create_rendered_chat_messages(messages)
        new_kwargs = dict(kwargs, messages=rendered_messages)

        start_time = time.time()
        completion = await super().create(**new_kwargs)

        if stream:
            is_self_contained = (
                    not self.pipeline_run and pipeline_slug
            )
            if is_self_contained:
                pipeline_run_id = str(uuid.uuid4())

            async def profiled_completion():
                modified_response = []
                async for value in completion:
                    if value and is_self_contained:
                        setattr(value, "pipelineRunId", pipeline_run_id)
                    modified_response.append(value)
                    yield value

                end_time = time.time()

                full_response = create_chat_completion_stream_response(
                    modified_response
                )

                elapsed_time = int((end_time - start_time) * 1000)

                pipeline_run = self.pipeline_run

                if is_self_contained:
                    pipeline = Pipeline(
                        slug=pipeline_slug,
                        **self.gentrace_config,
                    )

                    pipeline_run = PipelineRun(
                        pipeline=pipeline,
                        id=pipeline_run_id,
                        context=context,
                    )

                if pipeline_run:
                    pipeline_run.add_step_run(
                        OpenAICreateChatCompletionStepRun(
                            elapsed_time,
                            to_date_string(start_time),
                            to_date_string(end_time),
                            {
                                "messages": messages,
                                "user": user,
                                "contentInputs": content_inputs_array,
                            },
                            {
                                **model_params,
                                "contentTemplates": content_templates_array,
                            },
                            full_response,
                            context,
                        )
                    )

                    if is_self_contained:
                        pipeline_run.submit()

            return profiled_completion()

        end_time = time.time()

        elapsed_time = int((end_time - start_time) * 1000)

        pipeline_run = self.pipeline_run

        is_self_contained = not pipeline_run and pipeline_slug

        if is_self_contained:
            pipeline = Pipeline(
                slug=pipeline_slug,
                **self.gentrace_config
            )

            pipeline_run = PipelineRun(
                pipeline=pipeline,
                context=context,
            )

        if pipeline_run:
            pipeline_run.add_step_run(
                OpenAICreateChatCompletionStepRun(
                    elapsed_time,
                    to_date_string(start_time),
                    to_date_string(end_time),
                    {
                        "messages": messages,
                        "user": user,
                        "contentInputs": content_inputs_array,
                    },
                    {
                        **model_params,
                        "contentTemplates": content_templates_array,
                    },
                    completion.dict(),
                    context,
                )
            )

            if is_self_contained:
                submit_result = pipeline_run.submit()
                setattr(completion, "pipelineRunId",
                        submit_result["pipelineRunId"] if "pipelineRunId" in submit_result else None)

        return completion


class GentraceSyncOpenAI(OpenAI):
    pipeline_run: Optional[PipelineRun] = None

    def __init__(self, *args, **kwargs):
        # print all kwargs
        self.pipeline_run = kwargs.pop("pipeline_run", None)
        self.pipeline = kwargs.pop("pipeline", None)
        self.config = kwargs.pop("gentrace_config")

        super().__init__(*args, **kwargs)

        self.completions = GentraceSyncCompletions(client=self, gentrace_config=self.config,
                                                   pipeline=self.pipeline,
                                                   pipeline_run=self.pipeline_run)
        self.chat = ExtractedChat(self)
        self.chat.completions = GentraceSyncChatCompletions(client=self, gentrace_config=self.config,
                                                            pipeline=self.pipeline,
                                                            pipeline_run=self.pipeline_run)
        self.embeddings = GentraceSyncEmbeddings(client=self, gentrace_config=self.config,
                                                 pipeline=self.pipeline,
                                                 pipeline_run=self.pipeline_run)


class GentraceSyncEmbeddings(ExtractedEmbeddings):

    def __init__(self, *, gentrace_config: Dict[str, Any], client: OpenAI, pipeline: Optional[Pipeline] = None,
                 pipeline_run: Optional[PipelineRun] = None):
        super().__init__(client)

        self.pipeline = pipeline
        self.pipeline_run = pipeline_run
        self.gentrace_config = gentrace_config

    def create(self, *args, **kwargs):
        model = kwargs.get("model")
        pipeline_slug = kwargs.pop("pipeline_slug", None)
        context = kwargs.pop("gentrace", {})
        input_params = {k: v for k, v in kwargs.items() if k not in ["model"]}

        start_time = time.time()
        completion = super().create(**kwargs)
        end_time = time.time()

        elapsed_time = int((end_time - start_time) * 1000)

        pipeline_run = self.pipeline_run

        is_self_contained = not pipeline_run and pipeline_slug

        if is_self_contained:
            pipeline = Pipeline(
                slug=pipeline_slug,
                **self.gentrace_config,
            )

            pipeline_run = PipelineRun(
                pipeline=pipeline,
                context=context,
            )

        if pipeline_run:
            pipeline_run.add_step_run(
                OpenAICreateEmbeddingStepRun(
                    elapsed_time,
                    to_date_string(start_time),
                    to_date_string(end_time),
                    input_params,
                    {"model": model},
                    completion.dict(),
                    context,
                )
            )

            if is_self_contained:
                submit_result = pipeline_run.submit()
                setattr(completion, "pipelineRunId",
                        submit_result["pipelineRunId"] if "pipelineRunId" in submit_result else None)

        return completion


def create_chat_completion_stream_response(stream_list):
    final_response_string = ""
    for value in stream_list:
        if hasattr(value, "choices") and value.choices:
            first_choice = value.choices[0]
            if hasattr(first_choice, "text"):
                final_response_string += first_choice.text
            elif (
                    hasattr(first_choice, "delta")
                    and first_choice.delta
                    and hasattr(first_choice.delta, "content")
            ):
                final_response_string += first_choice.delta.content if first_choice.delta.content else ""
            elif (
                    hasattr(first_choice, "finish_reason")
                    and first_choice.finish_reason == "stop"
            ):
                break

    final_response = {
        "choices": [
            {
                "finish_reason": None,
                "index": 0,
                "message": {"content": final_response_string, "role": "assistant"},
            }
        ]
    }

    return final_response


class GentraceSyncChatCompletions(ExtractedChatCompletions):

    def __init__(self, *, gentrace_config: Dict[str, Any], client: OpenAI, pipeline: Optional[Pipeline] = None,
                 pipeline_run: Optional[PipelineRun] = None):
        super().__init__(client)

        self.pipeline = pipeline
        self.pipeline_run = pipeline_run
        self.gentrace_config = gentrace_config

    def create(self, *args, **kwargs):
        messages = kwargs.get("messages")
        user = kwargs.get("user")
        pipeline_slug = kwargs.pop("pipeline_slug", None)
        stream = kwargs.get("stream")
        context = kwargs.pop("gentrace", {})

        model_params = {
            k: v for k, v in kwargs.items() if k not in ["messages", "user"]
        }

        content_templates_array = [
            message["contentTemplate"] if "contentTemplate" in message else None
            for message in messages
        ]

        content_inputs_array = [
            message["contentInputs"] if "contentInputs" in message else None
            for message in messages
        ]

        if stream:
            rendered_messages = create_rendered_chat_messages(messages)
            new_kwargs = dict(kwargs, messages=rendered_messages)
            start_time = time.time()
            completion = super().create(*args, **new_kwargs)

            is_self_contained = (
                    not self.pipeline_run and pipeline_slug
            )
            if is_self_contained:
                pipeline_run_id = str(uuid.uuid4())

            def profiled_completion():
                modified_response = []
                for value in completion:
                    if value and is_self_contained:
                        setattr(value, "pipelineRunId", pipeline_run_id)
                    modified_response.append(value)
                    yield value

                end_time = time.time()

                elapsed_time = int((end_time - start_time) * 1000)

                pipeline_run = self.pipeline_run

                full_response = create_chat_completion_stream_response(
                    modified_response
                )
                if is_self_contained:
                    pipeline = Pipeline(
                        slug=pipeline_slug,
                        **self.gentrace_config,
                    )

                    pipeline_run = PipelineRun(
                        pipeline=pipeline, id=pipeline_run_id, context=context
                    )

                if pipeline_run:
                    pipeline_run.add_step_run(
                        OpenAICreateChatCompletionStepRun(
                            elapsed_time,
                            to_date_string(start_time),
                            to_date_string(end_time),
                            {
                                "messages": messages,
                                "user": user,
                                "contentInputs": content_inputs_array,
                            },
                            {
                                **model_params,
                                "contentTemplates": content_templates_array,
                            },
                            full_response,
                            context,
                        )
                    )

                    if is_self_contained:
                        pipeline_run.submit()

            return profiled_completion()

        rendered_messages = create_rendered_chat_messages(messages)

        new_kwargs = dict(kwargs, messages=rendered_messages)

        start_time = time.time()
        completion = super().create(*args, **new_kwargs)
        end_time = time.time()

        elapsed_time = int((end_time - start_time) * 1000)

        pipeline_run = self.pipeline_run

        is_self_contained = not pipeline_run and pipeline_slug
        if is_self_contained:
            pipeline = Pipeline(
                slug=pipeline_slug,
                **self.gentrace_config,
            )

            pipeline_run = PipelineRun(
                pipeline=pipeline,
                context=context,
            )

        if pipeline_run:
            pipeline_run.add_step_run(
                OpenAICreateChatCompletionStepRun(
                    elapsed_time,
                    to_date_string(start_time),
                    to_date_string(end_time),
                    {
                        "messages": messages,
                        "user": user,
                        "contentInputs": content_inputs_array,
                    },
                    {**model_params, "contentTemplates": content_templates_array},
                    completion.dict(),
                    context,
                )
            )

            if is_self_contained:
                submit_result = pipeline_run.submit()
                setattr(completion, "pipelineRunId",
                        submit_result["pipelineRunId"] if "pipelineRunId" in submit_result else None)

        return completion


def create_rendered_chat_messages(messages):
    new_messages = []
    for message in messages:
        new_message = copy.deepcopy(message)
        if "contentTemplate" in message and "contentInputs" in message:
            new_message["content"] = pystache.render(
                message["contentTemplate"], message["contentInputs"]
            )

            if "contentTemplate" in new_message:
                new_message.pop("contentTemplate", None)

            if "contentInputs" in new_message:
                new_message.pop("contentInputs", None)

        new_messages.append(new_message)

    return new_messages


class GentraceSyncCompletions(ExtractedCompletions):

    def __init__(self, *, gentrace_config: Dict[str, Any], client: OpenAI, pipeline: Optional[Pipeline] = None,
                 pipeline_run: Optional[PipelineRun] = None):
        super().__init__(client)

        self.pipeline = pipeline
        self.pipeline_run = pipeline_run
        self.gentrace_config = gentrace_config

    def create(self, *args, **kwargs):
        prompt_template = kwargs.get("prompt_template")
        prompt_inputs = kwargs.get("prompt_inputs")
        prompt = kwargs.get("prompt")
        pipeline_slug = kwargs.pop("pipeline_slug", None)
        context = kwargs.pop("gentrace", {})
        stream = kwargs.get("stream")
        base_completion_options = {
            k: v
            for k, v in kwargs.items()
            if k not in ["prompt_template", "prompt_inputs"]
        }

        if stream:
            rendered_prompt = prompt

            if prompt_template and prompt_inputs:
                rendered_prompt = pystache.render(prompt_template, prompt_inputs)
            else:
                # If no prompt template or inputs, then specify the prompt as the entire input
                prompt_inputs = prompt

            new_completion_options = {
                **base_completion_options,
                "prompt": rendered_prompt,
            }

            start_time = time.time()
            completion = super().create(*args, **new_completion_options)

            is_self_contained = (
                    not self.pipeline_run and pipeline_slug
            )
            if is_self_contained:
                pipeline_run_id = str(uuid.uuid4())

            def profiled_completion():
                modified_response = []
                for value in completion:
                    if value and is_self_contained:
                        setattr(value, "pipelineRunId", pipeline_run_id)
                    modified_response.append(value)
                    yield value

                end_time = time.time()

                full_response = create_completion_stream_response(modified_response)

                create_completion_step_run(
                    self.pipeline_run,
                    pipeline_slug,
                    self.gentrace_config,
                    start_time,
                    end_time,
                    base_completion_options,
                    prompt_template,
                    prompt_inputs,
                    completion,
                    full_response,
                    pipeline_run_id if is_self_contained else None,
                    stream,
                    context,
                )

            return profiled_completion()

        rendered_prompt = prompt

        if prompt_template and prompt_inputs:
            rendered_prompt = pystache.render(prompt_template, prompt_inputs)
        else:
            # If no prompt template or inputs, then specify the prompt as the entire input
            prompt_inputs = prompt

        new_completion_options = {
            **base_completion_options,
            "prompt": rendered_prompt,
        }

        start_time = time.time()
        completion = super().create(*args, **new_completion_options)

        end_time = time.time()

        create_completion_step_run(
            self.pipeline_run,
            pipeline_slug,
            self.gentrace_config,
            start_time,
            end_time,
            base_completion_options,
            prompt_template,
            prompt_inputs,
            completion,
            # Models need to converted from Pydantic to dict
            completion.dict(),
            None,
            stream,
            context,
        )

        return completion


def create_completion_step_run(
        pipeline_run: Optional[PipelineRun],
        pipeline_slug: str,
        gentrace_config: Dict[str, Any],
        start_time,
        end_time,
        base_completion_options,
        prompt_template,
        prompt_inputs,
        completion,
        completion_dict,
        pipeline_run_id: Optional[str] = None,
        stream=False,
        context={},
):
    elapsed_time = int((end_time - start_time) * 1000)

    user = base_completion_options.get("user")
    suffix = base_completion_options.get("suffix")
    partial_model_params = {
        k: v for k, v in base_completion_options.items() if k not in ["user", "suffix"]
    }

    inputs_dict = {"prompt": prompt_inputs}
    if user is not None:
        inputs_dict["user"] = user
    if suffix is not None:
        inputs_dict["suffix"] = suffix

    is_self_contained = not pipeline_run and pipeline_slug

    if is_self_contained:
        pipeline = Pipeline(
            slug=pipeline_slug,
            **gentrace_config,
        )

        pipeline_run = PipelineRun(
            pipeline=pipeline, id=pipeline_run_id, context=context
        )

    if pipeline_run:
        pipeline_run.add_step_run(
            OpenAICreateCompletionStepRun(
                elapsed_time,
                to_date_string(start_time),
                to_date_string(end_time),
                inputs_dict,
                {**partial_model_params, "promptTemplate": prompt_template},
                completion_dict,
                context,
            )
        )

        if is_self_contained:
            submit_result = pipeline_run.submit()

            if not stream:
                setattr(completion, "pipelineRunId",
                        submit_result["pipelineRunId"] if "pipelineRunId" in submit_result else None)


def create_completion_stream_response(stream_list):
    final_response_string = ""
    for value in stream_list:
        if hasattr(value, "choices") and value.choices:
            first_choice = value.choices[0]
            if hasattr(first_choice, "text"):
                final_response_string += first_choice.text
            elif (
                    hasattr(first_choice, "delta")
                    and first_choice.delta
                    and hasattr(first_choice.delta, "content")
            ):
                final_response_string += first_choice.delta.content
            elif (
                    hasattr(first_choice, "finish_reason")
                    and first_choice.finish_reason == "stop"
            ):
                break

    final_response = {
        "choices": [
            {
                "finish_reason": None,
                "index": 0,
                "logprobs": None,
                "text": final_response_string,
            }
        ]
    }

    return final_response


class OpenAICreateCompletionStepRun(StepRun):
    def __init__(
            self,
            elapsed_time,
            start_time,
            end_time,
            inputs,
            model_params,
            response,
            context,
    ):
        super().__init__(
            "openai",
            "openai_createCompletion",
            elapsed_time,
            start_time,
            end_time,
            inputs,
            model_params,
            response,
            context,
        )
        self.inputs = inputs
        self.model_params = model_params
        self.response = response


class OpenAICreateChatCompletionStepRun(StepRun):
    def __init__(
            self,
            elapsed_time,
            start_time,
            end_time,
            inputs,
            model_params,
            response,
            context,
    ):
        super().__init__(
            "openai",
            "openai_createChatCompletion",
            elapsed_time,
            start_time,
            end_time,
            inputs,
            model_params,
            response,
            context,
        )
        self.inputs = inputs
        self.model_params = model_params
        self.response = response


class OpenAICreateEmbeddingStepRun(StepRun):
    def __init__(
            self,
            elapsed_time,
            start_time,
            end_time,
            inputs,
            model_params,
            response,
            context,
    ):
        super().__init__(
            "openai",
            "openai_createEmbedding",
            elapsed_time,
            start_time,
            end_time,
            inputs,
            model_params,
            response,
            context,
        )
        self.inputs = inputs
        self.model_params = model_params
        self.response = response
